/*
 * Copyright (c) 1996, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gameengine.h"
#include "mapwindow.h"

const int map_sprite_width = 16;
const int map_sprite_height = 16;

FXDEFMAP(MapWindow) MapWindowMap[] = {
  FXMAPFUNC(SEL_COMMAND, MapWindow::ID_UPDATEMAP, MapWindow::onCmdUpdate)
};

FXIMPLEMENT(MapWindow, FXTopWindow, MapWindowMap, ARRAYNUMBER(MapWindowMap))

MapWindow::MapWindow()
{
}

MapWindow::MapWindow(FXApp *a, const FXString& name, GameEngine *e, const FXString& episode,
       		     FXIcon *icon, FXIcon *miniicon)
    : FXTopWindow(a, name, icon, miniicon, DECOR_CLOSE|DECOR_TITLE|DECOR_BORDER, 0,0,0,0, 0,0,0,0, 0,0)
{
	engine = e;

	for (int i = 0; i < num_sprites; ++i) {
		sprites[i] = new FXBMPImage(getApp(), 0, IMAGE_KEEP|IMAGE_OWNED|IMAGE_SHMI|IMAGE_SHMP,
					    sprite_width, sprite_height);
		sprites[i]->create();
	}
	map = new FXBMPImage(getApp(), 0, IMAGE_SHMI|IMAGE_SHMP,
 			     engine->width*map_sprite_width, engine->height*map_sprite_height);
	map->create();

	loadSprites(episode);

	FXVerticalFrame *contents =
        	new FXVerticalFrame(this, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 5,5,5,5, 5,5);
  	FXHorizontalFrame *mapframe =
   		new FXHorizontalFrame(contents, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_FILL_X,
   			0,0,0,0, 0,0,0,0, 0,0);
	mapview = new FXImageView(mapframe, map, 0, 0, LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,
				  0,0,engine->width*map_sprite_width,engine->height*map_sprite_height);
        new FXButton(contents, "&Update", 0, this, ID_UPDATEMAP,
	        BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,
         	0,0,0,0, 32,32,5,5);

	contents->setFocus();
}

MapWindow::~MapWindow()
{
	for (int i = 0; i < num_sprites; ++i)
		delete sprites[i];
	delete map;
}

void MapWindow::create()
{
  	FXTopWindow::create();
}

void MapWindow::loadSprites(const FXString& episode)
{
	FXBMPImage *img = new FXBMPImage(getApp(), 0);

	FXString filename = episode + PATHSEPSTRING "sprites.bmp";

	// Load it
  	FXFileStream stream;
  	if(!stream.open(filename, FXStreamLoad)) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "Cannot load %s", filename.text());
  		exit(1);
     	}

	img->loadPixels(stream);
	stream.close();
	img->create();

	if (img->getWidth() != 8*sprite_width || img->getHeight() != 6*sprite_height) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "The sprite bitmap %s must be 256x192", filename.text());
  		exit(1);
     	}

	for (int i = 0; i < num_sprites; ++i)
		sprites[i]->resize(sprite_width, sprite_height);

	// Load the sprites
	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x;
			FXDCWindow dc(sprites[i]);
		   	dc.setFunction(BLT_SRC);
			dc.drawImage(img, -x*sprite_width, -y*sprite_height);
		}

 	delete img;

	for (int i = 0; i < num_sprites; ++i) {
 		sprites[i]->restore();
		sprites[i]->scale(map_sprite_width, map_sprite_height);
	}
}

void MapWindow::updateSprites(const FXString& episode)
{
	loadSprites(episode);
}

void MapWindow::updateMap()
{
	int w, h, wdiff, hdiff;
	w = FXMIN(engine->width, 30)*map_sprite_width;
	h = FXMIN(engine->height, 30)*map_sprite_height;
	wdiff = w - map->getWidth();
 	hdiff = h - map->getHeight();
	if (wdiff != 0 || hdiff != 0) {
		map->resize(w, h);
		mapview->resize(w, h);
		resize(getWidth() + wdiff, getHeight() + hdiff);
  	}

	FXDCWindow dc(map);
   	dc.setFunction(BLT_SRC);
	for (int y = 0; y < engine->height; ++y)
		for (int x = 0; x < engine->width; ++x) {
			SpriteObj obj = engine->xy(x, y);
			switch (obj) {
			case Obj_Broken_Egg1:
			case Obj_Broken_Egg2:
			case Obj_Broken_Egg3:
				obj = Obj_Broken_Egg;
				break;
			case Obj_Falling_Egg:
				obj = Obj_Egg;
				break;
			default:;	// Avoid warning
			}
			dc.drawImage(sprites[obj], x*map_sprite_width, y*map_sprite_height);
		}

	mapview->update();
}

long MapWindow::onCmdUpdate(FXObject *, FXSelector, void *)
{
	updateMap();
	return 1;
}

