/*
 * Copyright (c) 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <algorithm>

#include "map.h"

Map::Map()
{
	data = 0;
	clear();
}

Map::~Map()
{
	clear();
}

void Map::clear()
{
	if (data) delete [] data;
	data = 0;
	width = height = 0;
	password = comments = FXString::null;
	deleteTransporters();
	ismodified = false;
	readonly = false;
	time_limit = 800;
}

void Map::copy(const Map *src)
{
 	width = src->width;
 	height = src->height;
 	ismodified = true;
  	readonly = src->readonly;
	password = src->password;
	comments = src->comments;
	time_limit = src->time_limit;

	data = new Sprite[width * height];

	for (int y = 0; y < height; ++y)
		for (int x = 0; x < width; ++x)
			xy(x, y) = src->xy(x, y);

  	typedef vector<Transporter *>::const_iterator I;
	for (I i = src->transporters.begin(); i != src->transporters.end(); ++i) {
		Transporter *t = *i;
		addTransporter(t->orig_x, t->orig_y, t->dest_x, t->dest_y);
	}
}

void Map::newMap(int w, int h)
{
	clear();
	width = w;
	height = h;

 	generatePassword();
	time_limit = 300;

	data = new Sprite[width * height];

	for (int y = 0; y < height; ++y)
		for (int x = 0; x < width; ++x)
			xy(x, y) = Sprite_Empty;
}

void Map::newDummyMap(int w, int h)
{
	newMap(w, h);
	xy(0, 0) = Sprite_Repton;
	xy(2, 0) = Sprite_Diamond;
	xy(2, 2) = Sprite_Bomb;
}

static FXString rot13(const FXString& str)
{
	FXString s = str;
 	for (int i = 0; i < s.length(); ++i) {
		FXuchar c = s[i];
		if (c >= 'a' && c <= 'z') {
 			c = c + 13;
 			if (c > 'z')
 				c = 'a' + c - 'z' - 1;
			s[i] = c;
		}
	}
	return s;
}

void Map::save(const FXString& filename)
{
	FILE *fout = fopen(filename.text(), "wb");
	if (!fout) {
//		if (showError) showError("Cannot open %s", filename);
	  	return;
  	}

	fprintf(fout, "%d %d %d %d %d\n",
 		width, height, transporters.size(), time_limit, readonly);
 	FXString pwd = rot13(password);
	fprintf(fout, "%s\n", pwd.text());

  	typedef vector<Transporter *>::const_iterator I;
	for (I i = transporters.begin(); i != transporters.end(); ++i) {
		Transporter *tr = *i;
		fprintf(fout, "%d %d %d %d\n", tr->orig_x, tr->orig_y, tr->dest_x, tr->dest_y);
	}

	for (int y = 0; y < height; ++y) {
		for (int x = 0; x < width; ++x)
			fputc(xy(x, y) + 'A', fout);
		fputc('\n', fout);
	}
	comments = comments.trim();
	fprintf(fout, "%s\n", (!comments.empty() ? comments.text() : "-"));

	fclose(fout);
}

void Map::load(const FXString& filename)
{
	clear();

	FILE *fin = fopen(filename.text(), "r");
	if (!fin) {
//		if (showError) showError("Cannot open %s", filename);
	  	return;
  	}

	char buf[1024];
	int num_transporters, ro;
	/* TODO (#1#): error checking */
	fgets(buf, 1024, fin);
	sscanf(buf, "%d %d %d %d %d", &width, &height, &num_transporters, &time_limit, &ro);
	readonly = ro;
	fgets(buf, 1024, fin);
 	password = rot13(FXString(buf).trim());

	while (num_transporters--) {
		int ox, oy, dx, dy;
		fgets(buf, 1024, fin);
		sscanf(buf, "%d %d %d %d\n", &ox, &oy, &dx, &dy);
		addTransporter(ox, oy, dx, dy);
	}

	data = new Sprite[width * height];

	for (int y = 0; y < height; ++y) {
		fgets(buf, 1024, fin);
		for (int x = 0; x < width; ++x) {
			char c = buf[x];
			xy(x, y) = (Sprite)(c - 'A');
		}
	}

	fgets(buf, 1024, fin);
 	comments = FXString(buf).trim();
	if (comments == "-")
		comments = FXString::null;

	fclose(fin);
}

void Map::resize(int w, int h, bool top, bool left)
{
#if 0
	Sprite *ndata = new Sprite[w*h];
	int xd, yd;
	if (w > width) {
		if (left)
  			xd = w - width;
	} else if (w < width) {
		if (left)
			xd = w - width;
 	}
	if (h > height) {
		if (top)
  			yd = h - height;
	} else if (h < height) {
		if (top)
			yd = h - height;
 	}
	for (int y = 0; y < h; ++y) {
		for (int x = 0; x < w; ++x) {
			Sprite s;
			if (x + xd < 0 || y + yd < 0 || x + xd >= width || y + yd >= height)
    				s = Sprite_Empty;
			   else
				s = xy(x + xd, y + yd);
			ndata[x + y*w] = s;
		}
	}
	delete data;
	width = w;
	height = h;

	for (I i = transporters.begin(); i != transporters.end(); ++i) {
		Transporter *t = *i;
		if (t->orig_x == x && t->orig_y == y)
			return t;
	}
#endif
}

void Map::addTransporter(int ox, int oy, int dx, int dy)
{
	Transporter *t = new Transporter(ox, oy, dx, dy);
	transporters.push_back(t);
}

Transporter *Map::findTransporter(int x, int y)
{
  	typedef vector<Transporter *>::const_iterator I;
	for (I i = transporters.begin(); i != transporters.end(); ++i) {
		Transporter *t = *i;
		if (t->orig_x == x && t->orig_y == y)
			return t;
	}
	return 0;
}

void Map::deleteTransporters()
{
  	typedef vector<Transporter *>::const_iterator I;
	for (I i = transporters.begin(); i != transporters.end(); ++i) {
		Transporter *t = *i;
		delete t;
	}
	transporters.clear();
}

bool Map::isModified() const
{
	return ismodified;
}

int Map::count(Sprite s) const
{
 	int i = 0;
 	for (int y = 0; y < height; ++y)
 		for (int x = 0; x < width; ++x)
    			if (xy(x, y) == s)
    				++i;
	return i;
}

static FXString newpwd(const char *templ)
{
	FXString s;
 	for (unsigned int i = 0; i < strlen(templ); ++i) {
 		unsigned char c = templ[i];
		if (c == 'X')
 			c = 'a' + rand() % 26;
 		s += c;
	}
	return s;
}

void Map::generatePassword()
{
	password = newpwd("XXXXXX");
}

Episode::Episode()
{
	isdataloaded = false;
	ismodified = false;
	isbroken = false;
}

Episode::~Episode()
{
	clear();
}

void Episode::clear()
{
	isdataloaded = false;
	ismodified = false;
	isbroken = false;
  	typedef vector<Map *>::const_iterator I;
	for (I i = maps.begin(); i != maps.end(); ++i) {
		Map *m = *i;
		delete m;
	}
	maps.clear();
	package = name = author = comments = path = FXString::null;
}

void Episode::appendMap(Map *map)
{
	maps.push_back(map);
}

void Episode::load(const FXString& epath)
{
	clear();
	
	path = epath;
	FXString fname = path + PATHSEPSTRING "info.fxr";

	FILE *fin = fopen(fname.text(), "r");
	if (!fin) {
//		if (showError) showError("Cannot open %s", filename);
	  	return;
  	}

	char buf[1024];
	int nolevels, broken;
	/* TODO (#1#): error checking */
	fgets(buf, 1024, fin); package = FXString(buf).trim();
	fgets(buf, 1024, fin); name = FXString(buf).trim();
	fgets(buf, 1024, fin); sscanf(buf, "%d %d", &nolevels, &broken); isbroken = broken;
     	fgets(buf, 1024, fin); author = FXString(buf).trim();
     	fgets(buf, 1024, fin); comments = FXString(buf).trim();
	fclose(fin);

	if (author == "-")
		author = FXString::null;
	if (comments == "-")
		comments = FXString::null;

	for (int i = 1; i <= nolevels; ++i) {
		fname = path + PATHSEPSTRING "map";
		fname += (char)(i + '0');
		fname += ".fxr";
		Map *m = new Map;
		m->load(fname);
		appendMap(m);
  	}

 	isdataloaded = true;
}

void Episode::save(const FXString& epath)
{
	FXString fname = epath + PATHSEPSTRING "info.fxr";

	FILE *fout = fopen(fname.text(), "wb");
	if (!fout) {
//		if (showError) showError("Cannot open %s", filename);
	  	return;
  	}

	author = author.trim();
	comments = comments.trim();

	fprintf(fout, "%s\n%s\n%d %d\n%s\n%s\n",
 		package.text(), name.text(), maps.size(), isbroken,
		(!author.empty() ? author.text() : "-"),
		(!comments.empty() ? comments.text() : "-"));
	fclose(fout);

	for (unsigned int i = 0; i < maps.size(); ++i) {
		fname = path + PATHSEPSTRING "map";
		fname += (char)(i + '0' + 1);
		fname += ".fxr";
		Map *m = maps[i];
		m->save(fname);
	}

 	ismodified = false;
	typedef vector<Map *>::const_iterator I;
	for (I i = maps.begin(); i != maps.end(); ++i) {
		Map *m = *i;
  		m->ismodified = false;
	}
}

bool Episode::isModified() const
{
	if (ismodified)
		return true;
  	typedef vector<Map *>::const_iterator I;
	for (I i = maps.begin(); i != maps.end(); ++i) {
		Map *m = *i;
  		if (m->isModified())
  			return true;
	}
	return false;
}

bool operator< (const Episode& e1, const Episode& e2)
{
	if (e1.package[0] == '[' && e2.package[0] != '[')
		return true;
	if (e1.package[0] != '[' && e2.package[0] == '[')
		return false;
	return (e1.package < e2.package) || (e1.package == e2.package && e1.name < e2.name);
}

void Episode::loadEpisodesList(vector<Episode>& v, const FXString& epath)
{
	DIR *dir = opendir(epath.text());
 	struct dirent *entry;

	if (!dir) {
		fprintf(stderr, "Cannot load episodes from %s\n", epath.text());
		return;
	}

	while ((entry = readdir(dir)) != NULL) {
		FXString path = epath;
		path += PATHSEP;
  		path += entry->d_name;
		FXString fname = path + PATHSEPSTRING "info.fxr";

		FILE *fin = fopen(fname.text(), "r");
		if (fin) {
     			char package[1024], epname[1024];
     			fgets(package, 1024, fin);
     			fgets(epname, 1024, fin);
		     	fclose(fin);

 			Episode ep;
    			ep.package = FXString(package).trim();
       			ep.name = FXString(epname).trim();
          		ep.path = path;
			v.push_back(ep);
    		}
	}
	closedir(dir);

	sort(v.begin(), v.end());
}

